"""
Input data _functions used to tidy semi-structured / structured numeric data.
"""

# Built-Ins
import logging
import os
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

# Third Party
import numpy as np
import pandas as pd

# Local Imports
from caf.brain.ml._functions._baseclasses import ValidateData

LOG = logging.getLogger(__name__)


class InitialDataProcessing:
    """Class for processing initial input data to the caf.brAIn prediction model."""

    def __init__(
        self,
        file_path: Path | None,
        folder_path: Path | None,
        output_path: Path | None,
        target_column: str | None,
        custom_index: list[str] | None,
        column_name_to_drop_rows: Optional[list[str]],
        value_in_row: list[str | float | int] | None,
        weight_column: str | None,
        categorical_features: list[str] | None,
        numerical_features: list[str] | None,
        classification_prediction: tuple[int, ...] | None,
        is_test_data: bool = False,
    ):
        """
        Parameters
        ----------

        file_path: Optional path to data to be used for modelling.
        folder_path: Optional path to folder of data to be used for
                     modelling.
        output_path: Path to output location.
        target_column: String column name of value to predict.
        custom_index: List of string column names to be used as an index.
                      Must be one value e.g. year if splitting data
                      into train and test via this column. Corresponds to
                      split_by_value in this case.
        column_name_to_drop_rows: List of string column names that
                                  contain values to drop.
        value_in_row: Corresponding values for column_name_to_drop_rows.
        weight_column: Optional string column value to be used as weight.
        categorical_features: List of string column names that are
                              categorical variables.
        numerical_features: List of string column names that are
                            continuous variables.
        classification_prediction: List of integers that correspond to the
                                   target column. The value(s) to predict
                                   in a classification problem.
        is_test_data: Set to False by default unless test data is
                      being processed. This is set to true when using
                      the main prediction function. If using separately
                      then set to True or False when calling the class.
        """

        self.file_path = file_path
        self.folder_path = folder_path
        self.output_path = output_path
        self.target_column = target_column
        self.custom_index = custom_index
        self.column_name_to_drop_rows = column_name_to_drop_rows
        self.value_in_row = value_in_row
        self.weight_column = weight_column
        self.categorical_features = categorical_features
        self.numerical_features = numerical_features
        self.classification_prediction = classification_prediction
        self.is_test_data = is_test_data

        self.df: pd.DataFrame = None
        self.dataframes: Dict[Any, pd.DataFrame] = {}

    # # # Data flow pipelines # # #
    def execute_pipeline(self, is_test_data: bool) -> Dict[str, pd.DataFrame]:
        """
        Run the InitialDataProcessing pipeline.

        Parameters
        ----------
        is_test_data: Dictates if data being processed is train or test.

        Returns
        -------
        Dictionary containing the finished processed dataframes
        """
        self.read_data()
        self.process_dataframes(is_test_data)
        self.validate_data(is_test_data)
        return self.dataframes

    def data_already_split_pipeline(self, is_test_data: bool) -> Dict[str, pd.DataFrame]:
        """
        Run the pipeline if train and test data is already split by the user
        or previous model runs.

        Parameters
        ----------
        is_test_data: Dictates if data being processed is train or test.

        Returns
        -------
        Dictionary containing the finished processed dataframes

        """
        self.process_dataframes(is_test_data)
        self.validate_data(is_test_data)
        return self.dataframes

    def read_data(self) -> None:
        """
        Read data from file or folder based on provided paths.

        Returns
        -------
        None
        Read data from file or folder based on provided paths.

        Returns
        -------
        None
        """
        if self.file_path:
            self.df = self.read_file(self.file_path)
            LOG.info("%s", self.df)
        elif self.folder_path:
            self.dataframes = self.read_folder(self.folder_path)
            LOG.info("%s", self.dataframes)
        else:
            LOG.error("Either file_path or folder_path must be provided")
            raise ValueError("Either file_path or folder_path must be provided")

    def process_dataframes(self, is_test_data: bool) -> None:
        """
        Process dataframe(s).

        Parameters
        ----------
        is_test_data: Dictates if data being processed is train or test.

        Returns
        -------
        None
        """
        LOG.info("Processing test if %s is True else training data", is_test_data)

        if len(self.dataframes) > 1:
            if self.custom_index is None:
                raise ValueError("Must provide custom index in order to process \
                                  multiple dataframes. Data processing outside of \
                                  caf.ml is advised.")
            self.df = pd.concat(self.dataframes, axis=0)

        if self.df.empty:
            raise ValueError(f"Dataframe {self.df} is empty")

        df = self.convert_to_dataframe(self.df)

        target_column_ = (
            []
            if is_test_data
            else ([self.target_column] if isinstance(self.target_column, str) else [])
        )
        weight_column_ = [self.weight_column] if isinstance(self.weight_column, str) else []
        custom_index = self.custom_index or []
        categorical_features = self.categorical_features or []
        numerical_features = self.numerical_features or []

        if self.numerical_features is None:
            columns_to_keep = (
                custom_index + categorical_features + target_column_ + weight_column_
            )
        elif self.categorical_features is None:
            columns_to_keep = (
                custom_index + numerical_features + target_column_ + weight_column_
            )
        else:
            columns_to_keep = (
                custom_index
                + categorical_features
                + numerical_features
                + target_column_
                + weight_column_
            )

        columns_to_keep = [col for col in columns_to_keep if col in df.columns]
        df = df[columns_to_keep]

        if is_test_data:
            LOG.info("Processing test data - skipping target column operations")
            df = self.function_remove_spaces(df)

            if self.column_name_to_drop_rows and self.value_in_row:
                df = self.drop_rows(df, self.column_name_to_drop_rows, self.value_in_row)

            df = self.handle_nans_and_duplicates(
                df, target_column=None, output_folder=self.output_path
            )
            df = df.apply(pd.to_numeric, errors="coerce")

            if self.custom_index:
                df = self.index_sorter(df, self.custom_index)

        else:
            df = self.function_remove_spaces(df)

            if self.column_name_to_drop_rows and self.value_in_row:
                df = self.drop_rows(df, self.column_name_to_drop_rows, self.value_in_row)

            if self.target_column not in df.columns:
                raise ValueError(
                    f"Target column '{self.target_column}' not found in training data"
                )

            df = self.handle_nans_and_duplicates(
                df, target_column=self.target_column, output_folder=self.output_path
            )

            df = self.numeric_transformation(df, self.target_column)

            if self.custom_index:
                df = self.index_sorter(df, self.custom_index)

            if self.classification_prediction:
                df = self.transform_target_column(
                    df,
                    target_column=self.target_column,
                    classification_prediction=self.classification_prediction,
                )

        df = self.convert_to_dataframe(df, columns=df.columns, index=df.index)

        filename = os.path.basename(self.file_path) if self.file_path else "processed_data"
        self.dataframes[filename] = df

    def validate_data(self, is_test_data) -> None:
        """
        Validate processed dataframes using the ValidateData class.
        Ensures data is in the format of the base class.

        Parameters
        ----------
        is_test_data: Dictates if data being processed is train or test.

        Returns
        -------
        None
        """
        LOG.info("Starting data validation")
        for _, df in self.dataframes.items():
            validator = ValidateData(
                dataframe=df, custom_index=self.custom_index, target_column=self.target_column
            )

            if is_test_data is False:
                validator.index_present()
                validator.target_column_present()
                validator.explanatory_data()
                validator.is_data_numeric()
                validator.data_correct_shape()
            else:
                LOG.info("Test data detected - skipping target column validation")
                validator.index_present()
                validator.explanatory_data()
                validator.is_data_numeric()

    # # # DATA PROCESSING METHODS # # #
    @staticmethod
    def read_file(file_path) -> pd.DataFrame:
        """
        Read in a file whilst trying all common encoding types and file types.

        Parameters
        ----------
        file_path: File path location.

        Returns
        -------
        Pandas dataframe of your input data.
        """
        file_extension = os.path.splitext(file_path)[1].lower()

        try:
            if file_extension == ".csv":
                encodings = ["utf-8", "latin-1", "iso-8859-1", "cp1252"]
                for encoding in encodings:
                    try:
                        df = pd.read_csv(file_path, encoding=encoding, low_memory=False)
                        return df
                    except (UnicodeDecodeError, pd.errors.ParserError):
                        continue
                raise ValueError(f"Unable to read CSV file with encodings: {encodings}")

            if file_extension in [".xlsx", ".xls"]:
                df = pd.read_excel(file_path)
                return df

            if file_extension == ".json":
                df = pd.read_json(file_path)
                return df

            raise ValueError(
                f"Unsupported file type: {file_extension}. Supported types: CSV, XLSX and JSON"
            )

        except Exception as e:
            LOG.error("Error reading file %s: %s", file_path, e)
            raise

    @staticmethod
    def read_folder(folder_path: Path) -> dict:
        """
        Create a dictionary of dataframes based on a path.
        #TODO MAKE IT READ IN OTHER FILE TYPES

        Parameters
        ----------
        folder_path: Path to a folder that contains csvs to be used as
                     input data.

        Returns
        -------
        Dictionary of Pandas dataframes based on your input data.
        """
        dataframes_dict = {}
        for file_name in os.listdir(folder_path):
            file_path = os.path.join(folder_path, file_name)
            if os.path.isfile(file_path) and file_name.endswith(".csv"):
                df = pd.read_csv(file_path, low_memory=False)
                dataframes_dict[file_name] = df
        return dataframes_dict

    @staticmethod
    def function_remove_spaces(df: pd.DataFrame) -> pd.DataFrame:
        """
        Remove any whitespace from column titles. A common problem in data.

        Parameters
        ----------
        df: Input dataframe.

        Returns
        -------
        Dataframe with tidy column titles.
        """
        df = df.map(lambda x: str(x).replace(" ", ""))
        return df

    @staticmethod
    def index_sorter(df: pd.DataFrame, custom_index: list[str]) -> pd.DataFrame:
        """
        Set columns in a dataframe as the index.

        Parameters
        ----------
        df: input dataframe.
        custom_index: List of strings that are columns in the dataframe.
                      These will be set as an index.

        Returns
        -------
        Input dataframe with a custom index.
        """
        for col in custom_index:
            if col not in df.columns:
                raise ValueError(f"Column '{col}' not found in DataFrame.")
        df.set_index(custom_index, inplace=True, verify_integrity=False)

        return df

    @staticmethod
    def drop_rows(
        df: pd.DataFrame,
        column_name_to_drop_rows: list[str],
        value_in_row: list[str | float | int],
    ) -> pd.DataFrame:
        """
        Remove specified rows from the input dataframe.

        Parameters
        ----------
        df: Input data.
        column_name_to_drop_rows: Column names where the problematic
                                  rows exist.
        value_in_row: The value inside the row that should be removed.

        Returns
        -------
        Dataframe with rows dropped.
        """
        df = df.astype(str)
        for col, val in zip(column_name_to_drop_rows, value_in_row):
            if col in df.columns:
                df = df[df[col] != val]
                LOG.info("Rows where %s is %s have been dropped.", col, val)
            else:
                LOG.warning("Column %s does not exist in the DataFrame.", col)
        return df

    @staticmethod
    def handle_nans_and_duplicates(
        dataframe: pd.DataFrame,
        target_column: str | None = None,
        output_folder: Path | None = None,
    ) -> pd.DataFrame:
        """
        Remove and export nans and duplicates.

        Parameters
        ----------
        dataframe: Input dataframe.
        target_column: Column in the dataframe specified by user.
        output_folder: Path to output folder.

        Returns
        -------
        Dataframe without nans and duplicates.
        """
        if target_column and dataframe[target_column].isna().any():
            raise ValueError(f"Target column '{target_column}' has NaN values. \
                               Please review the data.")

        rows_with_nans = dataframe[dataframe.isna().any(axis=1)]
        cleaned_dataframe = dataframe.dropna()

        if not rows_with_nans.empty and output_folder:
            nan_output_path = os.path.join(output_folder, "nans.csv")
            rows_with_nans.to_csv(nan_output_path, index=True)
            LOG.info("NaN rows exported to: %s", nan_output_path)
            LOG.info("NaN rows exported to: %s", nan_output_path)

        exact_duplicates = cleaned_dataframe[cleaned_dataframe.duplicated(keep=False)]

        if not exact_duplicates.empty:
            LOG.info("Exact duplicate rows found: %s", exact_duplicates)
            LOG.info("Exact duplicate rows found: %s", exact_duplicates)

            if output_folder:
                duplicates_output_path = os.path.join(output_folder, "exact_duplicates.csv")
                exact_duplicates.to_csv(duplicates_output_path, index=True)
                LOG.info("Exact duplicate rows exported to: %s", duplicates_output_path)
                LOG.info("Exact duplicate rows exported to: %s", duplicates_output_path)

            cleaned_dataframe = cleaned_dataframe.drop_duplicates()

        return cleaned_dataframe

    @staticmethod
    def numeric_transformation(data: pd.DataFrame, target_column: str | None) -> pd.DataFrame:
        """
        Ensure all data is numeric in a dataframe where applicable.

        Parameters
        ----------
        data: Input data.
        target_column: Column in the dataframe specified by user.

        Returns
        -------
        Pandas dataframe with numeric data where applicable.
        """
        if target_column not in data.columns:
            raise ValueError(
                "The target column is not in the dataframe. Please evaluate data."
            )

        data = data.apply(pd.to_numeric, errors="coerce")

        if not pd.to_numeric(data[target_column], errors="coerce").notna().all():
            data[target_column] = pd.to_numeric(data[target_column], errors="coerce")
            if not data[target_column].notna().all():
                raise ValueError("The target column could not be converted to numeric.")

        return data

    @staticmethod
    def convert_to_dataframe(
        data: Union[pd.DataFrame, List[Any], tuple, set, np.ndarray, pd.Series, dict],
        columns: Optional[List[str]] = None,
        index: Optional[List[Any]] = None,
    ) -> pd.DataFrame:
        """
        Convert data to a dataframe.

        Parameters
        ----------
        data: Input data of any regularly used formats.
        columns: List of column names to assign to dataframe.
        index: List of index labels to assign to index.

        Returns
        -------
        Dataframe with correct index and column titles.
        """
        try:
            if isinstance(data, pd.DataFrame):
                df = data
            elif isinstance(data, (list, tuple, set)):
                df = pd.DataFrame(data)
            elif isinstance(data, np.ndarray):
                df = pd.DataFrame(data)
            elif isinstance(data, pd.Series):
                df = pd.DataFrame({data.name: data})
            elif isinstance(data, dict):
                df = pd.DataFrame(data)
            else:
                LOG.error("Unsupported data type. Please provide a supported data type.")
                raise ValueError(
                    "Unsupported data type. Please provide a supported data type."
                )

            if columns is not None:
                df.columns = columns
            if index is not None:
                df.index = index

            return df
        except ValueError as n:
            LOG.error(
                "An error occurred during data conversion: %s. Data must be \
                       a standard Python, Numpy or Pandas datatype e.g. list, \
                       ndarray or series to be converted",
                n,
            )
            return None

    @staticmethod
    def transform_target_column(
        df: pd.DataFrame, target_column: str | None, classification_prediction: tuple[int, ...]
    ) -> pd.DataFrame:
        """
        Convert target column values to binary or multiclass.

        Parameters
        ----------
        df: Input dataframe
        target_column: String column name of value to predict.
        classification_prediction: List of integers that correspond to the
                                   target column. The value(s) to predict
                                   in a classification problem.

        Returns
        -------
        Dataframe with processed target column.
        """

        if (
            isinstance(classification_prediction, tuple)
            and len(classification_prediction) == 2
        ):
            LOG.info("Binary model selected for values %s", classification_prediction)
            LOG.info("Binary model selected for values %s", classification_prediction)
            df = df[df[target_column].isin(classification_prediction)]

        elif (
            isinstance(classification_prediction, tuple)
            and len(classification_prediction) == 3
        ):
            LOG.info("Multiclass model selected for values %s", classification_prediction)
            LOG.info("Multiclass model selected for values %s", classification_prediction)
            df = df[df[target_column].isin(classification_prediction)]

        df[target_column] = df[target_column].astype(int)
        unique_values = df[target_column].unique()
        LOG.info("Unique values in %s after transformation: %s", target_column, unique_values)
        LOG.info("Unique values in %s after transformation: %s", target_column, unique_values)

        return df



"""
Created on: 6/11/2024
Original author: Adil Zaheer
"""

# Built-Ins
import logging
import os
from pathlib import Path
from typing import List, Optional

# Third Party
import joblib
import pandas as pd
from pandas.core.dtypes.common import is_numeric_dtype
from sklearn.impute import SimpleImputer
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import StandardScaler

LOG = logging.getLogger(__name__)


def preprocess_numerical_data(
    df: pd.DataFrame,
    numerical_features: list[str],
    is_test_data: bool,
    numerical_pipeline_train=None,
    output_folder=None,
) -> tuple[pd.DataFrame, Optional[Pipeline]]:
    """
    Scales data via SciKitLearns standard scalar. Separates logic for train and
    test but ensures that the same transformations applied to train are applied
    to test if used in a full model run. If not used in a full model run,
    an output path must be provided to generate the training transformations
    which would then be applied to the test data when called again.

    Train example:
    is_test_data: False
    numerical_pipeline_train = None
    output_folder: Path

    Test example:
    is_test_data: True

    Parameters
    ----------
    df: Input dataframe.
    numerical_features: List of string column names that are
                               continuous variables.
    is_test_data: Bool. Used to dictate if test data is used or not.
    numerical_pipeline_train: Always None unless full model flow is
                                     running. Only is relevant to test data.
                                     If None then one of numerical_pipeline.pkl
                                     or scale_csv.csv needs to be available to
                                     apply transformations to train.
    output_folder: Path to output folder.

    Returns
    -------
    Dataframe of only the scaled numerical data set to the input datas index.
    Also, the numerical pipeline information when applicable.
    """

    if is_test_data:
        LOG.info("Processing test data - attempting to apply training transformations")
        numerical_pipeline = None
        method_used = None

        if numerical_pipeline_train is not None:
            try:
                numerical_pipeline = numerical_pipeline_train
                method_used = "in-memory pipeline"
            except ValueError as e:
                LOG.warning("Could not use in-memory pipeline: %s", str(e))

        if numerical_pipeline is None and output_folder is not None:
            numerical_pipeline_path_pkl = os.path.join(output_folder, "numerical_pipeline.pkl")
            try:
                numerical_pipeline = joblib.load(numerical_pipeline_path_pkl)
                method_used = "pickled pipeline"
            except ValueError as e:
                LOG.warning("Could not load pipeline from pickle: %s", str(e))

        if numerical_pipeline is None and output_folder is not None:
            scale_csv_path = os.path.join(output_folder, "scale_csv.csv")
            scale_df = pd.read_csv(scale_csv_path, index_col=0)

            scaler = StandardScaler()
            scaler.mean_ = scale_df.loc["mean"].values
            scaler.scale_ = scale_df.loc["std"].values
            scaler.var_ = scale_df.loc["var"].values

            try:
                numerical_pipeline = Pipeline(
                    [("imputer", SimpleImputer(strategy="median")), ("scaler", scaler)]
                )
                method_used = "CSV scaler values"
            except ValueError as e:
                LOG.warning("Could not load scaler values from CSV: %s", str(e))

        if numerical_pipeline is None:
            raise ValueError(
                "Unable to process test data: No transformation parameters available. \
                Please ensure either the numerical_pipeline_train parameter is provided, \
                or the output_folder contains 'numerical_pipeline.pkl' or 'scale_csv.csv'."
            )

        LOG.info(
            "Successfully applied training transformations to test data using %s", method_used
        )

        numerical_data = numerical_pipeline.transform(df[numerical_features])
        numerical_df = pd.DataFrame(numerical_data, columns=numerical_features, index=df.index)

        return numerical_df, None

    LOG.info("Processing training data - fitting new transformations")
    if output_folder is None:
        raise ValueError(
            "Output folder must be provided for training data to save transformations"
        )

    numerical_pipeline = Pipeline(
        [("imputer", SimpleImputer(strategy="median")), ("scaler", StandardScaler())]
    )

    numerical_data = numerical_pipeline.fit_transform(df[numerical_features])
    numerical_df = pd.DataFrame(numerical_data, columns=numerical_features, index=df.index)

    numerical_pipeline_path_pkl = os.path.join(output_folder, "numerical_pipeline.pkl")
    joblib.dump(numerical_pipeline, numerical_pipeline_path_pkl)
    LOG.info("Saved numerical pipeline to %s", numerical_pipeline_path_pkl)

    scale_csv_path = os.path.join(output_folder, "scale_csv.csv")
    scaler = numerical_pipeline.named_steps["scaler"]
    scale_df = pd.DataFrame(
        {"mean": scaler.mean_, "std": scaler.scale_, "var": scaler.var_},
        index=numerical_features,
    ).T
    scale_df.to_csv(scale_csv_path)
    LOG.info("Saved scaler values to %s", scale_csv_path)

    return numerical_df, numerical_pipeline


def preprocess_categorical_data(
    df: pd.DataFrame,
    categorical_features: list[str],
    sample_size_encode: bool | None,
    select_encode_values: bool | None,
    encode_values_to_drop: list[str] | None,
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Encodes categorical variables via a choice of methods. Standard
    encoding where the first in each category is dropped is default,
    see pandas.get_dummies documentation. Encoding via sample size or values
    set by the user is also possible.

    Parameters
    ----------
    df: Input dataframe.
    categorical_features: List of string column names that are
                          categorical variables.
    sample_size_encode: Optional bool. If true, the data will be split
                        based on sample size. Variables with the largest
                        sample size will be used as reference class.
    select_encode_values: Optional bool. If True, data is split based
                          on custom values set by the user. Corresponds
                          to encode_values_to_drop.
    encode_values_to_drop: If select_encode_values is True, then this
                           must be a list of strings the length of
                           categorical_features. Position one in the list
                           will link to the first variable provided in
                           categorical_features and so on.

    Returns
    -------
    categorical_df: Dataframe of encoded categorical data set
                    to the input datas index.
    drop_vals: Dataframe of columns removed during the encoding process.
    """
    df.columns = df.columns.astype(str)

    if sample_size_encode:
        categorical_df, drop_vals = sample_size_encode_(
            df=df, categorical_features=categorical_features
        )
        return categorical_df, drop_vals
    if select_encode_values:
        if not encode_values_to_drop:
            raise ValueError("")
        categorical_df, drop_vals = custom_sample_encode(
            df=df, categorical_features=categorical_features, drop_values=encode_values_to_drop
        )
        return categorical_df, drop_vals

    categorical_df = pd.get_dummies(
        df, columns=categorical_features, drop_first=True, dtype=float
    )
    categorical_df.columns = categorical_df.columns.str.replace(".0", "")

    cat_w_all_cols = pd.get_dummies(df, columns=categorical_features, dtype=float)
    cat_w_all_cols.columns = cat_w_all_cols.columns.str.replace(".0", "")

    extra_columns = cat_w_all_cols.columns.difference(categorical_df.columns)
    drop_vals = cat_w_all_cols[extra_columns]

    return categorical_df, drop_vals


def sample_size_encode_(
    df: pd.DataFrame, categorical_features: List[str]
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Encodes variables based on sample size. The value that appears most often
    in each of the categorical variables is used as the reference and therefore
    dropped during encoding.

    Parameters
    ----------
    df: Input dataframe.
    categorical_features: List of string column names that are
                          categorical variables.

    Returns
    -------
    Categorical_encoded: Dataframe of encoded categorical data.
    Dropped_df: Dataframe of columns removed during the encoding process.
    """
    modes = df[categorical_features].mode().iloc[0]

    modified_cat_features = []
    dropped_columns = []
    for feature in categorical_features:
        dummies = pd.get_dummies(
            df[feature], prefix=feature, prefix_sep="_", drop_first=False, dtype=float
        )

        mode_value = modes[feature]
        mode_column = f"{feature}_{mode_value}"

        if mode_column in dummies.columns:
            dropped_columns.append(mode_column)
            dummies = dummies.drop(columns=[mode_column])

        modified_cat_features.append(dummies)

    categorical_encoded = pd.concat(modified_cat_features, axis=1)

    # df_num = df.drop(columns=categorical_features)
    # df_encoded = pd.concat([df_num, categorical_encoded], axis=1)

    drop_df = pd.DataFrame({"Dropped_Columns": dropped_columns})
    new_columns = drop_df["Dropped_Columns"].tolist()
    dropped_df = pd.DataFrame(columns=new_columns)

    return categorical_encoded, dropped_df


def custom_sample_encode(
    df: pd.DataFrame, categorical_features: list[str], drop_values: list[str]
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """
    Encodes variables based on user specified values. The values should be
    specified in the order that the variables are listed in categorical_features.
    These will be the values dropped when encoded and therefore used as
    the reference.

    Parameters
    ----------
    df: Input dataframe.
    categorical_features: List of string column names that are
                          categorical variables.
    drop_values: List of strings the length of categorical_features.
                 Position one in the list will link to the first
                 variable provided in categorical_features and so on.

    Returns
    -------
    Categorical_encoded: Dataframe of encoded categorical data.
    Dropped_df: Dataframe of columns removed during the encoding process.
    """
    assert isinstance(categorical_features, list)
    assert isinstance(drop_values, list)
    if len(categorical_features) != len(drop_values):
        LOG.error("The number of categorical features must match the number of drop values")
        raise ValueError(
            "The number of categorical features must match the number of drop values"
        )

    modified_cat_features = []
    dropped_values = dict(zip(categorical_features, drop_values))
    dropped_columns = []

    for feature in categorical_features:
        dummies = pd.get_dummies(
            df[feature], prefix=feature, prefix_sep="_", drop_first=False, dtype=float
        )

        drop_value = dropped_values[feature]
        drop_column = f"{feature}_{drop_value}"

        if drop_column in dummies.columns:
            dropped_columns.append(drop_column)
            dummies = dummies.drop(columns=[drop_column])

        modified_cat_features.append(dummies)

    categorical_encoded = pd.concat(modified_cat_features, axis=1)
    # df_num = df.drop(columns=categorical_features)
    # df_encoded = pd.concat([df_num, categorical_encoded], axis=1)

    drop_df = pd.DataFrame({"Dropped_Columns": dropped_columns})
    new_columns = drop_df["Dropped_Columns"].tolist()
    dropped_df = pd.DataFrame(columns=new_columns)

    return categorical_encoded, dropped_df


def encode_test_data(
    test_df: pd.DataFrame,
    categorical_features: List[str],
    train_encoded: Optional[pd.DataFrame],
    target_column: str | None,
    weight_column: str | None,
    weight_df: pd.DataFrame,
) -> pd.DataFrame:
    """
    This encodes test data to match the training data.

    Parameters
    ----------
    test_df: Test data to be encoded.
    categorical_features: List of string column names that are
                          categorical variables.
    train_encoded: Encoded training data that the test data will match.
    target_column: Sting column name of value to predict.
    weight_column: Optional string column value to be used as weight.
    weight_df: The weight column in dataframe form to be added back
               to the encoded test data.

    Returns
    -------
    test_encoded: Encoded test data.
    """
    if train_encoded is None:
        raise ValueError("train_encoded is required")
    if target_column in train_encoded.columns:
        train_encoded = train_encoded.drop(columns=[target_column])
    if weight_column in train_encoded.columns:
        train_encoded = train_encoded.drop(columns=[weight_column])

    if target_column in test_df.columns:
        test_df = test_df.drop(columns=[target_column])

    test_encoded = pd.get_dummies(
        test_df, columns=categorical_features, prefix_sep="_", dtype=float
    )

    test_encoded = test_encoded.loc[:, test_encoded.columns.isin(train_encoded.columns)]

    if weight_df is not None:
        test_encoded[weight_column] = weight_df

    return test_encoded


def process_data_pipeline(
    df: pd.DataFrame,
    numerical_features: List[str] | None,
    categorical_features: List[str] | None,
    target_column: str | None,
    weight_column: str | None,
    sample_size_encode: bool | None,
    select_encode_values: bool | None,
    encode_values_to_drop: list[str] | None,
    train_encoded: Optional[pd.DataFrame],
    test_data: bool,
    numerical_pipeline,
    output_folder: Path,
) -> tuple[pd.DataFrame, Optional[pd.DataFrame], Optional[Pipeline]]:
    """
    Pipeline to process input data via encoding and scaling transformations.

    Parameters
    ----------
    df: Input dataframe.
    numerical_features: List of string column names that are
                        continuous variables.
    categorical_features: List of string column names that are
                          categorical variables.
    target_column: String column name of value to predict.
    weight_column: Optional string column value to be used as weight.
    sample_size_encode: Optional bool. If true, the data will be split
                        based on sample size. Variables with the largest
                        sample size will be used as reference class.
    select_encode_values: Optional bool. If True, data is split based
                          on custom values set by the user. Corresponds
                          to encode_values_to_drop.
    encode_values_to_drop: If select_encode_values is True, then this
                           must be a list of strings the length of
                           categorical_features. Position one in the list
                           will link to the first variable provided in
                           categorical_features and so on.
    train_encoded: Either None or an encoded and scaled train dataset.
                   This is to ensure that the corresponding test data
                   is encoding in the same way as the training data.
                   Columns in train and test must match for prediction.
    test_data: Bool, if True then the dataframe passed must be the test data.
    numerical_pipeline: This is the stored numerical pipline used on
                        the training data. Can be left as None if a full
                        model run is not being complete. In that case,
                        output_folder must contain the csv or pkl
                        version of numerical_pipeline.
    output_folder: Path to output folder

    Returns
    -------
    final_df: Scaled and encoded dataframe.
    drop_vals: Dataframe of columns removed during the encoding process.
    pipeline_out: stored transformation pipeline for continuous variables
    """
    if target_column is not None and target_column in df.columns:
        x = df.drop(columns=[target_column])
        y = df[target_column]
    else:
        x = df
        y = None

    final_df = None
    weight_df = None
    drop_vals = None
    pipeline_out = None

    if weight_column is not None and weight_column in df.columns:
        weight_df = df[weight_column]
        x = x.drop(columns=weight_column)

    # just numerical data
    if numerical_features is not None and categorical_features is None:
        if test_data:
            final_df, _ = preprocess_numerical_data(
                df=x,
                numerical_features=numerical_features,
                is_test_data=True,
                numerical_pipeline_train=numerical_pipeline,
                output_folder=output_folder,
            )

            if weight_df is not None:
                final_df[weight_column] = weight_df
        else:
            final_df, pipeline_out = preprocess_numerical_data(
                df=x,
                numerical_features=numerical_features,
                is_test_data=False,
                numerical_pipeline_train=numerical_pipeline,
                output_folder=output_folder,
            )

            if y is not None:
                final_df[target_column] = y

            if weight_df is not None:
                final_df[weight_column] = weight_df

    # just categorical data
    if categorical_features is not None and numerical_features is None:
        if test_data:
            final_df = encode_test_data(
                test_df=x,
                categorical_features=categorical_features,
                train_encoded=train_encoded,
                target_column=target_column,
                weight_column=weight_column,
                weight_df=weight_df,
            )

            if weight_df is not None:
                final_df[weight_column] = weight_df
        else:
            final_df, drop_vals = preprocess_categorical_data(
                df=x,
                categorical_features=categorical_features,
                sample_size_encode=sample_size_encode,
                select_encode_values=select_encode_values,
                encode_values_to_drop=encode_values_to_drop,
            )
            if y is not None:
                final_df[target_column] = y

            if weight_df is not None:
                final_df[weight_column] = weight_df

    # both categorical and numerical
    if numerical_features is not None and categorical_features is not None:
        x_cat = x.drop(columns=numerical_features)
        if test_data:
            numerical_df, _ = preprocess_numerical_data(
                df=x,
                numerical_features=numerical_features,
                is_test_data=True,
                numerical_pipeline_train=numerical_pipeline,
                output_folder=output_folder,
            )

            test_final = encode_test_data(
                test_df=x_cat,
                categorical_features=categorical_features,
                train_encoded=train_encoded,
                target_column=target_column,
                weight_column=weight_column,
                weight_df=weight_df,
            )

            final_df = pd.concat([numerical_df, test_final], axis=1)

            if weight_df is not None:
                final_df[weight_column] = weight_df
        else:
            numerical_df, pipeline_out = preprocess_numerical_data(
                df=x,
                numerical_features=numerical_features,
                is_test_data=False,
                numerical_pipeline_train=numerical_pipeline,
                output_folder=output_folder,
            )

            categorical_df, drop_vals = preprocess_categorical_data(
                df=x_cat,
                categorical_features=categorical_features,
                sample_size_encode=sample_size_encode,
                select_encode_values=select_encode_values,
                encode_values_to_drop=encode_values_to_drop,
            )
            final_df = pd.concat([numerical_df, categorical_df], axis=1)

            if y is not None:
                final_df[target_column] = y
                if not is_numeric_dtype(final_df[target_column]):
                    raise ValueError(f"Target column '{target_column}' must be numeric")

            if weight_df is not None:
                final_df[weight_column] = weight_df

    if drop_vals is not None and not drop_vals.empty:
        drop_vals.to_csv(os.path.join(output_folder, "dropped_encoding_vals.csv"), index=True)

    return final_df, drop_vals, pipeline_out



"""
Created on: 1/15/2025
Original author: Adil Zaheer
"""

# Built-Ins
import logging
import os.path
from pathlib import Path
from typing import Optional

# Third Party
import numpy as np
import pandas as pd
from sklearn.model_selection import train_test_split

# Local Imports
from caf.brain.ml._functions._ml_inputs import (
    DataClassificationInputs,
    Paths,
    TransformingInputDataInputs,
)
from caf.brain.ml._functions.process_data_functions.input_data import (
    InitialDataProcessing,
)

LOG = logging.getLogger(__name__)


def split_data(
    processed_dataframes: dict | pd.DataFrame,
    paths: Paths,
    data_classification: DataClassificationInputs,
    transforming_inputs: TransformingInputDataInputs,
    output_path,
) -> tuple[pd.DataFrame, pd.DataFrame, Optional[pd.DataFrame]]:
    """
    Function to split data into training and test if not already done by the
    user.

    Parameters
    ----------
    processed_dataframes: input dataframes inside a dictionary
    paths
    data_classification
    transforming_inputs
    output_path
    Returns
    -------
    Train, test and validate dataframes.

    """
    if isinstance(processed_dataframes, dict):
        df = pd.DataFrame.from_dict(processed_dataframes)
    else:
        df = processed_dataframes

    if transforming_inputs.split_by_value is not None:

        train, test, validate = split_by_column_value(
            df=df,
            index_columns=data_classification.custom_index,
            split_by_value=transforming_inputs.split_by_value,
            weight_column=data_classification.weight_column,
            target_column=data_classification.target_column,
            validation_path=paths.validation_path,
            output_path=output_path,
        )

        return train, test, validate

    train, test, validate = stratified_split_with_categories(
        df=df,
        categorical_features=data_classification.categorical_features,
        target_column=data_classification.target_column,
        weight_column=data_classification.weight_column,
        split_size=transforming_inputs.split_size,
        validation_path=paths.validation_path,
        index_columns=data_classification.custom_index,
        output_path=output_path,
    )
    return train, test, validate


def stratified_split_with_categories(
    df: pd.DataFrame,
    categorical_features: Optional[list[str]],
    target_column: Optional[str],
    weight_column: Optional[str],
    split_size: Optional[float],
    validation_path: Optional[Path],
    index_columns: Optional[list[str]],
    output_path: Path,
) -> tuple[pd.DataFrame, pd.DataFrame, Optional[pd.DataFrame]]:
    """
    Split data into train, test, and validate sets using stratification.

    Ensures all categories are represented at least once in both train and test.
    If split_size is None, 0.2 is used by default.

    Parameters
    ----------
    df: input dataframe.
    categorical_features: List of string column names that are
                          categorical variables.
    target_column: Sting column name of value to predict.
    weight_column: Optional string column value to be used as weight.
    split_size: Optional float e.g. 0.2. This would be the ratio to
                randomly split data into train and test. 0.2 is used
                if left as None.
    validation_path: Optional path to validation data if it exists.
                     This would need to correspond to the test data
                     created.
    index_columns: List of string column names to be used as an index.
    output_path: Path to output location.

    Returns
    -------
    Train, test and validate dataframes.
    """
    strat = pd.cut(df.iloc[:, 0], 4)
    train, test = train_test_split(
        df,
        test_size=split_size if split_size is not None else 0.2,
        random_state=42,
        stratify=strat,
    )

    if categorical_features is not None:
        missing_categories = {}
        for col in categorical_features:
            unique_vals = set(df[col].unique())
            test_vals = set(test[col].unique())
            missing = unique_vals - test_vals
            if missing:
                missing_categories[col] = missing

        if missing_categories:
            for col, missing_vals in missing_categories.items():
                for val in missing_vals:
                    missing_row = df[df[col] == val].iloc[[0]]
                    test = pd.concat([test, missing_row])
                    train = train[~train.index.isin(missing_row.index)]

    validate = None
    if weight_column in test.columns:
        test = test.drop(columns=weight_column)

    if target_column in test.columns:
        validate = pd.DataFrame({target_column: test[target_column]}, index=test.index)
        validate.to_csv(os.path.join(output_path, "validate.csv"), index=True)
        test = test.drop(columns=target_column)

    if validation_path is not None:
        validate = InitialDataProcessing.read_file(file_path=validation_path)
        if index_columns in validate.columns:
            validate = validate.set_index(index_columns)

    train.to_csv(os.path.join(output_path, "train.csv"), index=True)
    test.to_csv(os.path.join(output_path, "test.csv"), index=True)
    return train, test, validate


def split_by_column_value(
    df: pd.DataFrame,
    index_columns: Optional[list[str]],
    split_by_value: str,
    weight_column: Optional[str],
    target_column: Optional[str],
    validation_path: Optional[Path],
    output_path: Path,
) -> tuple[pd.DataFrame, pd.DataFrame, Optional[pd.DataFrame]]:
    """
    Split data into train, test, and validate sets by a specific column value.

    The value must correspond to the index column used. The index column must only
    be one column specified (no multi-index). For example, index_columns: year,
    split_by_value: '2019' means everything pre-2019 is training and everything
    post-2019 is test.

    Parameters
    ----------
    df: Input dataframe.
    index_columns: List of string column names to be used as an index.
                   Must be one value e.g. year if splitting data
                   into train and test via this column. Corresponds to
                   split_by_value in this case.
    split_by_value: Optional string that links to custom_index. The
                    value in the index column to split the data into
                    training and test.
    weight_column: Optional string column value to be used as weight.
    target_column: String column name of value to predict.
    validation_path: Optional path to validation data if it exists.
                     This would need to correspond to the test data created.
    output_path: Path to output location.

    Returns
    -------
    Train, test and validate dataframes.

    """
    if not index_columns or len(index_columns) != 1:
        raise ValueError("index_columns must contain exactly one column name for splitting")

    index_column = index_columns[0]

    if index_column not in df.index.names and index_column != df.index.name:
        raise KeyError(f"Index column '{index_column}' not found in DataFrame index")

    train = df.loc[df.index.get_level_values(index_column) <= int(split_by_value)]
    test = df.loc[df.index.get_level_values(index_column) > int(split_by_value)]

    validate = None
    if weight_column in test:
        test = test.drop(columns=weight_column)
    if target_column in test:
        validate = pd.DataFrame({target_column: test[target_column]}, index=test.index)
        validate.to_csv(os.path.join(output_path, "validate.csv"), index=True)
        test = test.drop(columns=target_column)

    train.to_csv(os.path.join(output_path, "train.csv"), index=True)
    test.to_csv(os.path.join(output_path, "test.csv"), index=True)

    if validation_path is not None:
        validate = InitialDataProcessing.read_file(file_path=validation_path)
        if index_column in validate.columns:
            validate = validate.set_index(index_column)

    return train, test, validate


def simple_train_test_split(
    df: pd.DataFrame, target_column: str | None, weight_column: str | None
):
    """
    Split data into train and test sets for model building.

    Parameters
    ----------
    df: Input dataframe.
    target_column: Name of the dependent (target) variable in the dataframe.
    weight_column: Name of the weight column in the dataframe.

    Returns
    -------
    x_train: Training features.
    x_test: Test features.
    y_train: Training target values.
    y_test: Test target values.
    x_train_weight: Weight values for the training set, if available.
    """
    if not target_column:
        raise ValueError("Please provide a target column. This should be a \
                          column title passed as a string.")
    x = df.drop(columns=[target_column])
    y = df[target_column]
    x_train, x_test, y_train, y_test = train_test_split(x, y, test_size=0.35, random_state=42)

    x_train_weight = None
    if weight_column in df.columns:
        x_train_weight = x_train[weight_column]
        x_train = x_train.drop(columns=weight_column)
        x_test = x_test.drop(columns=weight_column)

    return x_train, x_test, y_train, y_test, x_train_weight


def sample_data(
    x: pd.DataFrame,
    y: pd.DataFrame,
    weight: np.ndarray | None = None,
    is_time_series: bool = False,
):
    """
    Take a sample of data whilst maintaining temporal nature of data if
    applicable.

    Parameters
    ----------
    x: Pandas dataframe of explanatory variable data.
    y: Pandas dataframe of target variable data.
    weight: Numpy array of weight column if applicable.
    is_time_series: True if data is time series.

    Returns
    -------
    x_sample: Sample of explanatory variables or just the input x.
    y_sample: Sample of target variable or just the input y.
    weight_sample: Sample of weight variable or the input weight if weight is
                   provided otherwise None.
    """
    if not y.index.equals(x.index):
        raise ValueError("X and y must have identical indices.")

    n_rows = len(x)
    use_sampling = n_rows > 500000

    if use_sampling:
        LOG.warning(
            "Dataset has %d rows. Sampling 500,000 rows for memory efficiency.", n_rows
        )

        if is_time_series:
            sample_positions = np.arange(n_rows - 500000, n_rows)
        else:
            sample_positions = np.random.RandomState(42).choice(n_rows, 500000, replace=False)

        x_sample = x.iloc[sample_positions]
        y_sample = y.iloc[sample_positions]
        weight_sample = weight[sample_positions] if weight is not None else None

    else:
        x_sample = x
        y_sample = y
        weight_sample = weight

    return x_sample, y_sample, weight_sample



"""
Main function for model selection capabilities.
"""

# Built-Ins
import logging
from pathlib import Path

# Third Party
import pandas as pd
from sklearn.base import BaseEstimator

# Local Imports
from caf.brain.ml._functions._ml_inputs import (
    DataClassificationInputs,
    ModellingInputs,
    Paths,
    TransformingInputDataInputs,
)
from caf.brain.ml._functions.model_selection.functions import select_model
from caf.brain.ml._functions.process_data_functions.main import (
    main_input_data,
)

LOG = logging.getLogger(__name__)


def main_model_selection(
    data_classification: DataClassificationInputs,
    transforming_inputs: TransformingInputDataInputs,
    modelling: ModellingInputs,
    output: Path,
    paths: Paths,
    train: pd.DataFrame = None,
) -> BaseEstimator:
    """
    Function to automatically score and rank algorithms from the Models
    class which can be used in machine learning modelling.

    Parameters
    ----------
    data_classification: Data classification inputs from the PredictionModelInputs
                         class. These inputs help define and outline the
                         structure of the input data. See
                         caf/brain/ml/main_models/prediction_model/_ml_inputs.py
                         for available options.
    transforming_inputs: Transforming inputs from the PredictionModelInputs
                         class. These inputs dictate how the data is transformed
                         for machine learning modelling. See
                         caf/brain/ml/main_models/prediction_model/_ml_inputs.py
                         for available options.
    modelling: Modelling inputs from the PredictionModelInputs
               class. These inputs control the machine learning modelling
               pipeline and _functions. See
               caf/brain/ml/main_models/prediction_model/_ml_inputs.py
               for available options.
    paths: Path inputs from the PredictionModelInputs class. These inputs
           define paths to external files. See
           caf/brain/ml/main_models/prediction_model/_ml_inputs.py
           for available options.
    output: Path to output location.
    train: Processed input data split into train subset.

    Returns
    -------
    Selected_model: The best performing SciKitLearn model based on the user
                    provided list. This model is not initialised and ready
                    for further use.
    """
    if train is None:
        LOG.info(
            "Train is none so data is being read in from PredictionModelInputs.Paths.file_path"
        )

        data_dict, _, _ = main_input_data(
            output_path=output,
            paths=paths,
            data_classification=data_classification,
            transforming_inputs=transforming_inputs,
        )
        LOG.info("Data successfully read in, processed and validated")
        train = pd.DataFrame.from_dict(data_dict["train_scaled"])
    if not modelling.model_choice:
        raise ValueError("Please provide at least one model from the Models \
                          class.")

    if not isinstance(modelling.model_choice, list):
        model = [modelling.model_choice]
    else:
        model = modelling.model_choice

    if len(model) == 1:
        selected_model = model[0].get_model()

    elif len(model) > 1:
        LOG.info("Beginning model evaluation.")
        selected_model = select_model(
            train=train,
            target_column=data_classification.target_column,
            weight_column=data_classification.weight_column,
            models_to_test=model,
            classification_prediction=transforming_inputs.classification_prediction,
            output_folder=output,
            is_time_series=data_classification.is_time_series,
        )
    else:
        LOG.error("Model incorrectly provided or not provided at all \
                          Provide a valid model(s) from the Models Enum class.")
        raise ValueError("Model incorrectly provided or not provided at all \
                          Provide a valid model(s) from the Models Enum class.")

    return selected_model



"""
Created on: 15/1/2025
Original author: Adil Zaheer
"""

# Built-Ins
import argparse
import os
from pathlib import Path

# Third Party
from caf.toolkit import LogHelper, ToolDetails

# Local Imports
from caf.brain.ml._functions._ml_inputs import PredictionModelInputs
from caf.brain.ml._functions.prediction_model_main import main


def _custom_load_yaml(config_path: Path) -> PredictionModelInputs:
    """
    Loads the YAML configuration file for the caf.brAIn full machine learning
    pipline and returns its contents as a dictionary.

    If no path is provided, defaults to 'brain.yml' which you can put in the
    current working directory.

    Parameters
    ----------
    config_path: Path to the YAML configuration file.

    Returns
    -------
    Parsed contents of the YAML file as a dictionary.
    """
    if config_path is None:
        config_path = Path.cwd() / "brain.yml"
    if not config_path.exists():
        raise FileNotFoundError(f"No config file found at {config_path}")

    config_data = PredictionModelInputs.load_yaml(config_path)

    return config_data


def model_setup():
    """
    Function to set up logging files, output folders and input data
    for the caf.brAIn full machine learning pipline.
    """
    parser = argparse.ArgumentParser(
        description="Run caf.brAIn prediction model.",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument(
        "--config",
        type=Path,
        default=Path("brain.yml"),
        help=(
            "Path to YAML config file. You should use docs/config.rst \n"
            "as guidance and examples/brain.yml as a template.\n"
        ),
    )
    args = parser.parse_args()
    params = _custom_load_yaml(args.config)

    output_path = params.paths.output_path / "output"
    if not output_path.is_dir():
        os.makedirs(output_path)

    path = output_path / "log_file.log"
    details = ToolDetails("caf.brAIn Prediction Model", "1.0.0")

    with LogHelper("caf.brain", details, console=True, log_file=path):
        main(params, output_path)


if __name__ == "__main__":
    model_setup()



"""
Created on: 1/16/2025
Original author: Adil Zaheer
"""

# Built-Ins
import gc
import logging
import os
from pathlib import Path

# Third Party
import joblib
import numpy as np
import pandas as pd
from sklearn.ensemble import (
    ExtraTreesClassifier,
    ExtraTreesRegressor,
    RandomForestClassifier,
    RandomForestRegressor,
)
from sklearn.linear_model import LinearRegression
from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor

# Local Imports
from caf.brain.ml._functions._ml_inputs import (
    ModelGrids,
    get_model_grid_from_type,
)
from caf.brain.ml._functions.feature_selection.functions import get_cv_class
from caf.brain.ml._functions.process_data_functions.split_data_into_ttv import (
    sample_data,
)

LOG = logging.getLogger(__name__)


def select_param(
    train_final: pd.DataFrame,
    target_column: str | None,
    model_instance,
    model_name,
    classification_prediction: tuple[int, ...] | None,
    cv: str | None,
    weight_column: str | None,
    output_folder: Path,
    is_time_series: bool | None,
):
    """
    Hyperparameter optimisation based on the ModelGrids Enum class.

    Parameters
    ----------
    train_final: Dataframe of final training data post feature selection.
    target_column: String column name of value to predict.
    model_instance: Initialised model algorithm from Models enum class.
    model_name: List or one algorithm to use as the base of the model.
                Available algorithms can be seen in _ml_inputs.py.
    classification_prediction: List of integers that correspond to the
                               target column. The value(s) to predict
                               in a classification problem.
    cv: Cross validation method passed as a string. Any popular
        SciKitlearn methods are suitable with KFold being default if
        left as None.
    weight_column: Optional string column value to be used as weight.
    output_folder: Path to output location.
    is_time_series: If true then data must be time series. Time series
                    based characteristics are taken into consideration
                    during function execution.

    Returns
    -------
    best_model: Fitted final model for prediction on unseen (test) data.

    """
    model_filename = os.path.join(output_folder, "final_model.pkl")
    filename = "final_model.pkl"
    model_path = Path(output_folder) / filename
    if model_path.exists():
        final_model = joblib.load(model_path)
        return final_model

    if not target_column:
        raise ValueError("Please provide a target column for hyperparameter \
                          optimisation. This is a column title passed as a string.")
    x = train_final.drop(columns=[target_column] + ([weight_column] if weight_column else []))
    y = train_final[target_column]
    weight = train_final[weight_column].values.flatten() if weight_column else None
    cv = get_cv_class(cv_method=cv, splits=None, repeats=None, is_time_series=is_time_series)

    x, y, weight = sample_data(x=x, y=y, weight=weight, is_time_series=is_time_series)

    if isinstance(model_name, list):
        if len(model_name) == 1:
            param_grid = ModelGrids.get_grid(model_name[0])
        else:
            param_grid = get_model_grid_from_type(type(model_instance))
    else:
        param_grid = ModelGrids.get_grid(model_name)

    if isinstance(model_instance, LinearRegression):
        LOG.info(
            "No hyperparameters in Linear Regression. Skipping hyperparameter optimisation."
        )
        return model_instance

    if classification_prediction is not None:
        if isinstance(
            model_instance,
            (RandomForestClassifier, ExtraTreesClassifier, DecisionTreeClassifier),
        ):
            if isinstance(model_instance, DecisionTreeClassifier):
                model_instance.set_params(max_depth=10)
                best_params = rand_search(
                    model_instance=model_instance,
                    param_grid=param_grid,
                    cv=cv,
                    scoring="accuracy",
                    n_jobs=-1,
                    weight=weight,
                    x=x,
                    y=y,
                )
            else:
                model_instance.set_params(n_estimators=10, n_jobs=-1)
                best_params = perform_grid_search(
                    model_instance=model_instance,
                    param_grid=param_grid,
                    cv=cv,
                    scoring="accuracy",
                    n_jobs=-1,
                    weight=weight,
                    x=x,
                    y=y,
                )
        else:
            best_params = perform_grid_search(
                model_instance=model_instance,
                param_grid=param_grid,
                cv=cv,
                scoring="accuracy",
                n_jobs=-1,
                weight=weight,
                x=x,
                y=y,
            )

    else:
        if isinstance(
            model_instance, (RandomForestRegressor, ExtraTreesRegressor, DecisionTreeRegressor)
        ):
            if isinstance(model_instance, DecisionTreeRegressor):
                model_instance.set_params(max_depth=10)
                best_params = rand_search(
                    model_instance=model_instance,
                    param_grid=param_grid,
                    cv=cv,
                    scoring="r2",
                    n_jobs=-1,
                    weight=weight,
                    x=x,
                    y=y,
                )
            else:
                model_instance.set_params(n_estimators=10, n_jobs=-1)
                best_params = perform_grid_search(
                    model_instance=model_instance,
                    param_grid=param_grid,
                    cv=cv,
                    scoring="r2",
                    n_jobs=-1,
                    weight=weight,
                    x=x,
                    y=y,
                )
        else:
            best_params = perform_grid_search(
                model_instance=model_instance,
                param_grid=param_grid,
                cv=cv,
                scoring="r2",
                n_jobs=-1,
                weight=weight,
                x=x,
                y=y,
            )

    best_model = model_instance.set_params(**best_params)
    best_model.fit(x, y, sample_weight=weight)

    joblib.dump(best_model, model_filename)
    LOG.info("Best model saved here: %s", model_filename)
    # coeffs
    if hasattr(best_model, "coef_"):
        coefficients = best_model.coef_
        coefficients = np.squeeze(coefficients)

        if coefficients.ndim == 1:
            coeff_df = pd.DataFrame({"Feature": x.columns, "Coefficient": coefficients})
        else:
            coeff_df = pd.DataFrame(coefficients.T, index=x.columns)

        coeff_df.to_csv(
            os.path.join(output_folder, "final_model_coefficients.csv"), index=False
        )

    return best_model


def rand_search(
    model_instance,
    param_grid,
    cv,
    scoring: str,
    n_jobs: int,
    weight: pd.Series,
    x: pd.DataFrame,
    y: pd.DataFrame,
):
    """
    Helper function to conduct randomised search of hyperparameters.

    Parameters
    ----------
    model_instance: Initialised model algorithm from Models enum class.
    param_grid: Instance of Enum class ModelGrids. Grid of hyperparameters.
    cv: Cross validation method passed as a string. Any popular
        SciKitlearn methods are suitable with KFold being default if
        left as None.
    scoring: Scoring method used when evaluating hyperparameters.
    n_jobs: Integer to represent number of cores used to evaluate hyperparameters.
    weight: Weight values in series form.
    x: Training data split into explanatory variables only.
    y: Training data split only into the target variable.

    Returns
    -------
    best_params: best hyperparameters found.

    """
    LOG.info("Performing randomised search for hyperparameter optimisation.")
    random_search = RandomizedSearchCV(
        model_instance,
        param_grid,
        cv=cv,
        scoring=scoring,
        verbose=2,
        n_jobs=n_jobs,
        n_iter=10,
        return_train_score=False,
        pre_dispatch="1*n_jobs",
    )
    gc.collect()
    random_search.fit(x, y, sample_weight=weight)
    best_params = random_search.best_params_
    LOG.info("Best parameters for model are: %s", best_params)
    LOG.info("CV results: %s", random_search.cv_results_)
    return best_params


def perform_grid_search(
    model_instance,
    param_grid,
    cv,
    scoring: str,
    n_jobs: int,
    weight: pd.Series,
    x: pd.DataFrame,
    y: pd.DataFrame,
):
    """
    Helper function to conduct grid search of hyperparameters.

    Parameters
    ----------
    model_instance: Initialised model algorithm from Models enum class.
    param_grid: Instance of Enum class ModelGrids. Grid of hyperparameters
    cv: Cross validation method passed as a string. Any popular
        SciKitlearn methods are suitable with KFold being default if
        left as None.
    scoring: Scoring method used when evaluating hyperparameters.
    n_jobs: Integer to represent number of cores used to evaluate hyperparameters.
    weight: Weight values in series form.
    x: Training data split into explanatory variables only.
    y: Training data split only into the target variable.

    Returns
    -------
    best_params: best hyperparameters found.
    """
    LOG.info("Performing grid search for hyperparameter optimisation.")
    grid_search = GridSearchCV(
        model_instance,
        param_grid,
        cv=cv,
        scoring=scoring,
        verbose=2,
        n_jobs=n_jobs,
        return_train_score=False,
        pre_dispatch="1*n_jobs",
    )
    gc.collect()
    grid_search.fit(x, y, sample_weight=weight)
    best_params = grid_search.best_params_
    LOG.info("Best parameters for model are: %s", best_params)
    LOG.info("CV results: %s", grid_search.cv_results_)
    return best_params
